package gatewayapi.sample;

import java.util.Collection;
import java.util.Optional;
import java.util.Scanner;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import com.google.gson.Gson;
import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.mmbnetworks.gatewayapi.ConnectionInfo;
import com.mmbnetworks.gatewayapi.ConnectionInfo.ConnectionType;
import com.mmbnetworks.gatewayapi.GatewayClient;
import com.mmbnetworks.gatewayapi.PropertyCommandData;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.LightDevice;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.gatewayapi.entity.ThermostatDevice;
import com.mmbnetworks.gatewayapi.entity.ZigbeeDevice;
import com.mmbnetworks.gatewayapi.event.device.DeviceEvent;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventHandler;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventStatus;
import com.mmbnetworks.gatewayapi.exception.GatewayConnectionException;
import com.mmbnetworks.gatewayapi.exception.InvalidInputException;
import com.mmbnetworks.rapidconnectconnections.SerialUtil;
import com.mmbnetworks.rapidconnectdevice.zcl.ZigBee;
import com.mmbnetworks.rapidconnectdevice.zcl.cluster.FanControl;
import com.mmbnetworks.rapidconnectdevice.zcl.cluster.Thermostat;
import com.mmbnetworks.rapidconnectdevice.zigbee.LatencyInfo;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.JSONUtils;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.MessageType;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.ZCLMulticastMessage.ZCLMulticastMessageBuilder;

import ch.qos.logback.classic.Level;

/**
 * This sample application demonstrates how one might use the GatewayClient to connect to a known
 * network interface and interact with devices on it.
 * <p>
 * Usage: SampleApp portName
 * <p>
 * Example 1: java -jar SampleApp.jar
 * Example 2: java -jar SampleApp.jar /dev/ttyUSB0
 * Example 3: java -jar SampleApp.jar /dev/ttyUSB0 /path/to/config/file.xml
 * 
 */
public class SimpleCli implements DeviceEventHandler {
    
    private static final String[] commands = {
            "==== General Commands ====",
            " h - help",
            " toggleLog - turn logging on/off",
            " q - quit",
            " v - get version",
            " n - network information",
            " d - list devices and their device ids",
            "==== Gateway Client ====",
            " f - form network",
            " p - permit join (30 sec)",
            " c - close network permit join window",
            " l - dissolve network",
            " getDevicesOfType - get a list of devices of a certain type",
            "==== Device Commands ====",
            " getDeviceType - get the device type of a device",
            " getProperties - get properties for a selected device id. Will perform a network request",
            " getCachedProperties - get last known properties for a selected device id. Will be empty if there are no last cached properties",
            " getProtocolProperties - gets the properties of the device in the format of the underlying protocol (i.e. zigbee). Will perform a network request",
            " getCachedProtocolProperties - gets the last known properties of the device in the format of the underlying protocol (i.e. zigbee). Will be empty if there are no last cached properties",
            " getProperty - get a specific property. Will perform a network request",
            " getCachedProperty - get the last known value of a specific property. Will be empty if there is no last known value. Will throw exception if property doesn't exist",
            " updateProperty - update a specific property for a selected device id",
            " discoverAll - perform a full discovery of all properties on a device over the network (warning: can be slow)",
            " getConnectedGatewayId - get the ID of the Gateway that this Device is connected to",
            " bind - bind a source to a destination",
            " unbind - unbind a source to a destination",
            " enableReporting - enable reporting on properties using default parameters",
            " addSwitchHandler - map a switch device to a group id",
            " addToGroup - add a device to a group",
            " removeFromGroup - remove a device from a group",
            " kickDevice - kick a remote device off the network",
            " removePropertyListeners - remove property update listeners for the selected device",
            " ping - ping a remote device to get its round trip latency",
            "",
            "==== Light Commands ====",
            " on",
            " off",
            " moveToLevel - change the dimmable light level",
            " readLevel - read the dimmable light level",
            "",
            "==== Thermostat Commands ====",
            " readSystemMode - read the system mode, i.e. heat, cool, auto, off, etc.",
            " readFanMode - read the fan mode, i.e. off, auto, etc.",
            " readRunningMode - read the running mode of the thermostat",
            " changeSystemMode - change system mode, i.e. heat, cool, auto, off, etc.",
            " changeFanMode - change the fan mode, i.e. off, auto, etc.",
            " readOccupiedCoolingSetpoint - read the cooling setpoint",
            " readOccupiedHeatingSetpoint - read the heating setpoint",
            " setOccupiedCoolingSetpoint - set the cooling setpoint to a specific temperature",
            " setOccupiedHeatingSetpoint - set the heating setpoint to a specific temperature",
            " adjustOccupiedSetpoint - adjust the heating or cooling setpoint up or down by a specific amount",
            "",
            "==== Zigbee Commands ====",
            " readAttribute - read a zigbee attribute, knowing the cluster id and attribute id",
            " writeAttribute - write a zigbee attribute, knowing the cluster id, attribute id, data type, and byte values",
            " sendZclCommand - send a zcl command, knowing the command id, payload bytes, and cluster id",
            "",
    };
    
    private static ch.qos.logback.classic.Logger rootLogger;
    private static Level originalLogLevel;
    private static Level currentLogLevel;

    public static void main(String[] args) {

        // setup scanner to read from std in
        final Scanner scanner = new Scanner(System.in, "UTF-8");
        
        // get a GatewayClient instance
        final GatewayClient gw = getGatewayClient(args, scanner);
        if (gw == null) {
            cleanUp(null, scanner);
            return;
        }

        setupLogging();
        
        printCommands(commands);

        runMainLoop(gw, scanner);

        cleanUp(gw, scanner);

        logger("Exited SampleApp");
    }
    
    private static GatewayClient getGatewayClient(String[] args, Scanner scanner) {
        GatewayClient gw = null;
        
        // if no args, prompt user to enter serial port and use default config file
        if (args.length == 0) {
            while (true) {
                logger("Enter a serial port (i.e. /dev/ttyUSB0), or 'q' to exit:");
                final String input = scanner.nextLine();

                if ("q".equalsIgnoreCase(input)) {
                    logger("Quitting...");
                    scanner.close();
                    return null;

                } else {
                    try {
                        gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, input), new SimpleCli());
                        break;
                    } catch (GatewayConnectionException e) { 
                        logger("Could not connect or create GatewayClient, try again");
                    }
                }
            }

        // if one arg provided, assume it's the port and use default config file
        } else if (args.length == 1) {
            final String devicePort = args[0];
            try {
                gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, devicePort), new SimpleCli());
            } catch (GatewayConnectionException e) { 
                logger("Could not connect or create GatewayClient with port: " + devicePort);
                scanner.close();
                return null;
            }

        // if two args provided, assume first arg is the port and second arg is the config file path
        } else if (args.length == 2) {
            final String devicePort = args[0];
            final String configFilePath = args[1];
            try {
                gw = new GatewayClient(new ConnectionInfo(ConnectionType.ZIGBEE_UART, devicePort), new SimpleCli(), configFilePath);
            } catch (GatewayConnectionException e) { 
                logger("Could not connect or create GatewayClient with port: " + devicePort);
                scanner.close();
                return null;
            }

        } else {
            logger("Invalid number of arguments.");
            scanner.close();
            return null;
        }
        
        return gw;
    }
    
    private static void setupLogging() {
        rootLogger = (ch.qos.logback.classic.Logger) org.slf4j.LoggerFactory.getLogger(ch.qos.logback.classic.Logger.ROOT_LOGGER_NAME);
        originalLogLevel = rootLogger.getLevel();
        currentLogLevel = originalLogLevel;
        rootLogger.setLevel(Level.OFF);
        currentLogLevel = Level.OFF;
    }
    
    private static void cleanUp(GatewayClient gw, Scanner scanner) {
        if (gw != null) {
            final boolean shutdown = gw.shutdown();
            if (!shutdown) {
                logger("Error shutting down gateway client, check logs");
            }    
        }
        
        scanner.close();
    }
    
    private static void logger(final String msg) {
        System.out.println(msg);
    }
    
    private static void printCommands(final String[] cmds) {
        logger("Available Commands: ");
        for (String s : cmds) {
            logger(s);
        }
    }
    
    private static void printDeviceEvent(DeviceEventStatus eventStatus, String deviceId, String deviceType, String deviceCategory) {
        StringBuilder sb = new StringBuilder();
        sb.append("Device Event: ").append(eventStatus.toString()).append(", ")
          .append("Device Id: ").append(deviceId).append(", ")
          .append("Device Type: ").append(deviceType).append(", ")
          .append("Device Category: ").append(deviceCategory);
        System.out.println(sb.toString());
    }
    
    private static String listEnumNames(Enum[] values) {
        final StringBuilder modes = new StringBuilder();
        boolean first = true;
        for (Enum e : values) {
            if (!first) { modes.append(", "); } else { first = !first; } 
            modes.append(e.toString().replaceAll("_", " "));
        }
        return modes.toString();
    }

    private static void runMainLoop(GatewayClient gw, Scanner scanner) {
        while (true) {
            try {
                logger("Enter a command: ");
                final String input = scanner.nextLine();
                
                if ("h".equalsIgnoreCase(input)) {
                    printCommands(commands);
    
                } else if ("toggleLog".equalsIgnoreCase(input)) {
                    if (currentLogLevel != Level.OFF) { 
                        rootLogger.setLevel(Level.OFF);
                        currentLogLevel = Level.OFF;
                    } else {
                        rootLogger.setLevel(originalLogLevel);
                        currentLogLevel = originalLogLevel;
                    }
    
                } else if ("v".equalsIgnoreCase(input)) {
                    logger("API Version = " + GatewayClient.getApiVersion());
    
                } else if ("q".equalsIgnoreCase(input)) {
                    break;
    
                } else if ("n".equalsIgnoreCase(input)) {
                    for (ConnectionInfo c: gw.getConnectionInfo()) {
                        gw.getNetworkInfo(c).thenAccept( networkInformation -> {
                            logger("Information for connection " + c.getValue() + " is " + networkInformation);
                        });   
                    }
    
                } else if ("f".equalsIgnoreCase(input)) {
                    for (ConnectionInfo c: gw.getConnectionInfo()) {
                        gw.createNetwork(c).thenAccept( networkInformation -> {
                            logger("Information for connection " + c.getValue() + " is " + networkInformation);
                        });
                    }
    
                } else if ("p".equalsIgnoreCase(input)) {
                    for (ConnectionInfo c: gw.getConnectionInfo()) {
                        gw.scanForDevices(c, 30).thenAccept( networkInformation -> {
                            logger("Information for connection " + c.getValue() + " is " + networkInformation);
                        });
                    }
    
                } else if ("c".equalsIgnoreCase(input)) {
                    for (ConnectionInfo c: gw.getConnectionInfo()) {
                        gw.scanForDevices(c, 0).thenAccept( networkInformation -> {
                            logger("Information for connection " + c.getValue() + " is " + networkInformation);
                        });
                    }
    
                } else if ("l".equalsIgnoreCase(input)) {
                    for (ConnectionInfo c: gw.getConnectionInfo()) {
                        gw.dissolveNetwork(c).thenAccept( networkInformation -> {
                            logger("Information for connection " + c.getValue() + " is " + networkInformation);
                        });
                    }
    
                } else if ("d".equalsIgnoreCase(input)) {
                    final Collection<Device> devices = gw.getDevices();
                    if (devices.size() == 0) {
                        logger("No devices found on the network");
    
                    } else {
                        int i = 1;
                        for (Device d : devices) {
                            StringBuilder sb = new StringBuilder();
                            sb.append(i++).append(") ")
                              .append("Device Id: ").append(d.getID())
                              .append(", Type: ").append(d.getDeviceType().getType())
                              .append(", Category: ").append(d.getDeviceType().getCategory());
                            
                            logger(sb.toString());
                        }
                    }
    
                } else if ("getDevicesOfType".equalsIgnoreCase(input)) {
                    logger("Enter a device type (i.e. dimmable light, thermostat, occupancy sensor, combined interface, etc): ");
                    final String typeStr = scanner.nextLine();
                    final Collection<Device> devices = gw.getDevices( d -> d.getDeviceType().getType().equalsIgnoreCase(typeStr) );
                    int i = 1;
                    for (Device d : devices) {
                        StringBuilder sb = new StringBuilder();
                        sb.append(i++).append(") ")
                          .append("Device Id: ").append(d.getID())
                          .append(", Type: ").append(d.getDeviceType().getType())
                          .append(", Category: ").append(d.getDeviceType().getCategory());
                        
                        logger(sb.toString());
                    }
    
                } else if ("getDeviceType".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    
                    StringBuilder sb = new StringBuilder();
                    sb.append("Device Id: ").append(d.getID())
                    .append(", Type: ").append(d.getDeviceType().getType())
                    .append(", Category: ").append(d.getDeviceType().getCategory());
                    
                    logger(sb.toString());
    
                } else if ("getCachedProperties".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
    
                    final Collection<Property> properties = d.getCachedProperties();
                    if (properties.isEmpty()) {
                        logger("No properties found");
                    } else {
                        for (Property p : properties) {
                            logger(p.toString());
                        }   
                    }
    
                } else if ("getProperties".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) {
                        logger("Device not found");
                        continue;
                    }
                    final CompletableFuture<Collection<Property>> result = d.getProperties();
                    logger("Retrieving properties...");
                    try {
                        Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                        for (Property p : properties) {
                            logger(p.toString());
                        }
                    } catch (InterruptedException | ExecutionException | TimeoutException e) {
                        logger("Error, failed to retrieve updated properties");
                        e.printStackTrace();
                    }
                } else if ("getCachedProtocolProperties".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
    
                    final Collection<Property> properties = d.getCachedProtocolProperties();
                    if (properties.isEmpty()) {
                        logger("No properties found");
                    } else {
                        for (Property p : properties) {
                            logger(p.toString());
                        }
                    }
                    
                } else if ("getProtocolProperties".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) {
                        logger("Device not found");
                        continue;
                    }
                    final CompletableFuture<Collection<Property>> result = d.getProtocolProperties();
                    logger("Retrieving protocol properties...");
                    try {
                        Collection<Property> properties = result.get(30, TimeUnit.SECONDS);
                        for (Property p : properties) {
                            logger(p.toString());
                        }
                    } catch (InterruptedException | ExecutionException | TimeoutException e) {
                        logger("Error, failed to retrieve updated properties: " + e.getMessage());
                        e.printStackTrace();
                    }
                } else if ("getProperty".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device device = gw.getDevice(idStr);
                    if (device == null) { logger("Device \'" + idStr + "\' not found\n"); continue; }

                    logger("Enter a property name: ");
                    final String propertyName = scanner.nextLine();
                    
                    final CompletableFuture<Property> p = device.getProperty(propertyName);
                    try {
                        logger("Retrieving property \'" + propertyName + "\'");
                        final Property property = p.get(30, TimeUnit.SECONDS);
                        logger(property.toString());
                    } catch (InterruptedException | ExecutionException | TimeoutException e) {
                        logger("Error, failed to retrieve updated property: " + e.getMessage());
                    }
    
                } else if ("getCachedProperty".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
    
                    logger("Enter a property name: ");
                    final String propertyName = scanner.nextLine();
    
                    final Device device = gw.getDevice(idStr);
                    if (device == null) { logger("device not found"); continue; }
    
                    final Optional<Property> pOpt = device.getCachedProperty(propertyName);
                    if (!pOpt.isPresent()) {
                        logger("Property \'" + propertyName + "\' not found");
                        continue;
                    }
                    Property property = pOpt.get();
                    logger(property.toString());
                } else if ("enableReporting".equalsIgnoreCase(input)) {
                    logger("Enter a device id:");
                    String idStr = scanner.nextLine();
                    final Device device = gw.getDevice(idStr);
                    if (device == null) { logger("device not found"); continue; }
    
                    try {
                        final CompletableFuture<String> f = device.enableDefaultReporting();
                        final String result = f.get(30, TimeUnit.SECONDS);
                        logger("Status for command: " + result);
                    } catch (IllegalStateException | InterruptedException | ExecutionException | TimeoutException e) {
                        logger(e.getMessage());
                    }
                
                } else if ("bind".equalsIgnoreCase(input)) {
                    try {
                        logger("Enter a source id:");
                        final String source = scanner.nextLine();
                        final Device sourceDevice = gw.getDevice(source);
                        if (sourceDevice == null) {
                            logger("Invalid Source Device: \'" + source + "\'");
                            continue;
                        }
                        logger("Enter a destination id:");
                        final String destination = scanner.nextLine();
                        final Device destinationDevice = gw.getDevice(destination);
                        if (destinationDevice == null) {
                            logger("Invalid Destination Device: \'" + destination + "\'");
                            continue;
                        }
                        logger("Enter a Property name to bind:");
                        final String property = scanner.nextLine();
                        sourceDevice.bindDevice(destinationDevice.getID(), property)
                                    .whenComplete((result, error) -> {
                                        if (error != null) {
                                            logger("error calling bind: " + error.getMessage());
                                        }
                                        logger("bind device result: " + result);
                                    });
    
                    } catch (Exception ex) {
                        logger("error: " + ex.getMessage());
                    }
    
                } else if ("getConnectedGatewayId".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device device = gw.getDevice(idStr);
                    if (device == null) { logger("device not found"); continue; }
    
                    logger("Connected GatewayId: " + device.getConnectedGatewayId().orElse("n/a"));
    
                } else if ("addToGroup".equalsIgnoreCase(input)) {
                    
                    logger("Enter a device id:");
                    final String source = scanner.nextLine();
                    final Device d = gw.getDevice(source);
                    if (d == null) { logger("no device found"); continue; }
                    
                    logger("Enter a group id (i.e. 1-10):");
                    final String groupIdString = scanner.nextLine();
                    int groupId = 0;
                    try {
                        groupId = Integer.parseInt(groupIdString);
                    } catch (NumberFormatException e) {
                        logger("invalid number");
                        continue;
                    }
                    
                    final CompletableFuture<String> response = d.addToGroup(groupId);
                    response.thenAccept( msg -> logger(msg) );
    
                } else if ("removeFromGroup".equalsIgnoreCase(input)) {
                    
                    logger("Enter a device id:");
                    final String source = scanner.nextLine();
                    final Device d = gw.getDevice(source);
                    if (d == null) { logger("no device found"); continue; }
                    
                    logger("Enter a group id (i.e. 1-10):");
                    final String groupIdString = scanner.nextLine();
                    int groupId = 0;
                    try {
                        groupId = Integer.parseInt(groupIdString);
                    } catch (NumberFormatException e) {
                        logger("invalid number");
                        continue;
                    }
                    
                    final CompletableFuture<String> response = d.removeFromGroup(groupId);
                    response.thenAccept( msg -> logger(msg) );
    
                } else if ("addSwitchHandler".equalsIgnoreCase(input)) {
                    
                    logger("Enter a source id (i.e. the switch):");
                    final String source = scanner.nextLine();
                    final Device sourceDevice = gw.getDevice(source);
                    if (sourceDevice == null) { logger("no device found"); continue; }
                    
                    logger("Enter a Group Id to re-broadcast switch message to:");
                    final String groupIdString = scanner.nextLine();
                    try {
                        Integer.parseInt(groupIdString);
                    } catch (NumberFormatException e) {
                        logger("invalid number");
                        continue;
                    }
                    final GatewayClient finalGwClient = gw;
                    
                    sourceDevice.addProtocolHandler( (device,jsonMsg) -> {
                        logger("Rebroadcast handler: " + jsonMsg); 
                        try {
                            final JsonParser parser = new JsonParser();
                            final JsonObject jsonObject = parser.parse(jsonMsg).getAsJsonObject();
                            final JsonObject message = jsonObject.get("message").getAsJsonObject();
    
                            // send multicast protocol message
                            final ZCLMulticastMessageBuilder builder = new ZCLMulticastMessageBuilder();
                            builder.setGatewayAPIVersion(GatewayClient.getApiVersion());
                            builder.setProtocolName("zigbee");
                            builder.setProtocolVersion(3);
                            builder.setMessageType(MessageType.ZCL_MULTICAST);
                            builder.setGroupID(groupIdString);
                            builder.setClusterID(message.get("clusterId").getAsString());
                            builder.setFrameControl(ZigBee.FrameConstants.FRAME_CONTROL_CLIENT_TO_SERVER_CLUSTER_CMD.toBitmap8());
                            builder.setCommandID(message.get("commandId").getAsString());
                            // just send on the first available network interface; we're assuming we're only running on one for now
                            finalGwClient.sendProtocolMessage(
                                    finalGwClient.getConnectionInfo().iterator().next(),
                                    builder.build().toJson());
    
                        } catch (Exception ex) {
                            ex.printStackTrace();
                        }
                    });
                    
                } else if ("unbind".equalsIgnoreCase(input)) {
                    try {
                        logger("Enter a source id:");
                        final String source = scanner.nextLine();
                        final Device sourceDevice = gw.getDevice(source);
                        if (sourceDevice == null) {
                            logger("Invalid Source Device: \'" + source + "\'");
                            continue;
                        }
                        logger("Enter a destination id:");
                        final String destination = scanner.nextLine();
                        final Device destinationDevice = gw.getDevice(destination);
                        if (destinationDevice == null) {
                            logger("Invalid Destination Device: \'" + destination + "\'");
                            continue;
                        }
                        logger("Enter a Property name to unbind:");
                        final String property = scanner.nextLine();
                        sourceDevice.unbindDevice(destinationDevice.getID(), property)
                                    .whenComplete((result, error) -> {
                                        if (error != null) {
                                            logger("error calling unbind: " + error.getMessage());
                                        }
                                        logger("unbind device result: " + result);
                                    });
    
                    } catch (Exception ex) {
                        logger("error: " + ex.getMessage());
                    }
    
                } else if ("kickDevice".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    d.leaveNetwork();
    
                } else if ("ping".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    try {
                        String latencyInfoString = d.ping().get(15, TimeUnit.SECONDS);
                        LatencyInfo latencyInfo = new Gson().fromJson(latencyInfoString, LatencyInfo.class);
                        logger("[status=" + latencyInfo.status.getLabel().toUpperCase() + 
                               ", directOrIndirect=" + latencyInfo.directOrIndirect.getLabel().toUpperCase() + 
                               ", latency=" + latencyInfo.latency + "ms" + 
                               ", lastHopLQI=" + latencyInfo.lastHopLQI + 
                               ", lastHopRSSI=" + latencyInfo.lastHopRSSI + "dBm]");
                    } catch (Exception e) {
                        logger("Error, failed to retrieve latency information");
                        e.printStackTrace();
                    }

                } else if ("removePropertyListeners".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    d.removeAllPropertyUpdateHandlers();
    
                } else if ("updateProperty".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    
                    final Device d = gw.getDevice(idStr);
                    if (d == null) {
                        logger("Device not found");
                        continue;
                    }
                    
                    logger("Enter a property name: ");
                    final String propertyId = scanner.nextLine();
                    
                    logger("Enter a value: ");
                    final String writeValue = scanner.nextLine();
    
                    Property property = new Property(propertyId, "na", writeValue);
                    d.updateProperty(property).getFuture().whenComplete( (result, throwable) -> {
                        if (throwable != null) {
                            logger("error: " + throwable.getMessage());
                        } else if (result != null) {
                            logger("result: " + result.getValue());    
                        }
                    });
    
                } else if ("discoverAll".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
    
                    CompletableFuture<Collection<Property>> properties = d.discoverAllProperties();
                    logger("Performing discovery of all properties (can take a while)...");
                    properties.thenAccept( collection -> {
                        for (Property p : collection) {
                            logger("Discovered: " + p.toString());
                        }
                    });
    
                } else if ("on".equalsIgnoreCase(input) || "off".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("device not found"); continue; }
                    if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                    
                    final LightDevice light = (LightDevice) d;
                    CompletableFuture<Property> f = null;
                    if ("on".equalsIgnoreCase(input)) { f = light.on().getFuture(); }
                    else if ("off".equalsIgnoreCase(input)) { f = light.off().getFuture(); }
                    
                    // print errors if they happen
                    if (f != null) {
                        f.whenComplete( (result, throwable) -> {
                            if (throwable != null) {
                                logger("error: " + throwable.getMessage());
                            }
                        });
                    }
    
                } else if ("moveToLevel".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("device not found"); continue; }
                    if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
                    
                    logger("Enter a level (as percent, 0-100): ");
                    final String levelPercentageStr = scanner.nextLine();
                    int levelPercentage = 0;
                    try {
                        levelPercentage = Integer.parseInt(levelPercentageStr);
                    } catch (NumberFormatException e) {
                        logger("invalid number");
                        continue;
                    }
                    
                    final LightDevice light = (LightDevice) d;
                    light.moveToLevel(levelPercentage);
    
                } else if ("readLevel".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("device not found"); continue; }
                    if (!(d instanceof LightDevice)) { logger("device is not a light"); continue; }
    
                    final LightDevice light = (LightDevice) d;
                    try {
                        String level = light.readLevel().get(5, TimeUnit.SECONDS).getValue();
                        logger("Light Level for " + light.getID() + ": " + level);
                    } catch (InterruptedException | ExecutionException | TimeoutException e) {}
    
                } else if ("sendZigBeeMessage".equalsIgnoreCase(input)) {
                    logger("Enter Device ID:");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) {
                        logger("device not found");
                        continue;
                    }
                    logger("enter json message:");
                    String json = scanner.nextLine();
                    try {
                        CompletableFuture<String> future = d.sendProtocolMessage(json);
                        String result = future.get(30, TimeUnit.SECONDS);
                        System.err.println("Result: " + result);
                    } catch (InvalidInputException e) {
                        logger("Invalid JSON: "+ json);
                        e.printStackTrace();
                    } catch (InterruptedException | ExecutionException | TimeoutException e) {
                        e.printStackTrace();
                    }
                } else if (("changeSystemMode".equalsIgnoreCase(input))
                        || ("changeFanMode".equalsIgnoreCase(input))
                        || ("readSystemMode".equalsIgnoreCase(input))
                        || ("readFanMode".equalsIgnoreCase(input))
                        || ("readRunningMode".equalsIgnoreCase(input))
                        || ("readOccupiedCoolingSetpoint".equalsIgnoreCase(input))
                        || ("readOccupiedHeatingSetpoint".equalsIgnoreCase(input))
                        || ("setOccupiedCoolingSetpoint".equalsIgnoreCase(input))
                        || ("setOccupiedHeatingSetpoint".equalsIgnoreCase(input))
                        || ("adjustOccupiedSetpoint".equalsIgnoreCase(input))) {
    
                    // get the device
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null ) { logger("Device not found"); continue; }
                    if (!(d instanceof ThermostatDevice)) { logger("device is not a thermostat"); continue; }
    
                    // parse the command
                    final ThermostatDevice device = (ThermostatDevice) d;
                    PropertyCommandData command = null;
                    CompletableFuture<Property> propertyFuture = null;
                    if ("readSystemMode".equalsIgnoreCase(input)) {
                        propertyFuture = device.readSystemMode();
    
                    } else if ("changeSystemMode".equalsIgnoreCase(input)) {
                        final StringBuilder modes = new StringBuilder();
                        modes.append("enter mode (");
                        modes.append(listEnumNames(Thermostat.SystemModeEnum.values()));
                        modes.append("): ");
                        logger(modes.toString());
                        final String value = scanner.nextLine();
                        for (Thermostat.SystemModeEnum systemMode : Thermostat.SystemModeEnum.values()) {
                            if (systemMode.name().replace("_","").equalsIgnoreCase(value.replace(" ", ""))) {
                               command = device.changeSystemMode(systemMode);
                               break;
                            }
                        }
    
                    } else if ("adjustOccupiedSetpoint".equalsIgnoreCase(input)) {
                        CompletableFuture<String> setpointCommand = null;
                        final StringBuilder modes = new StringBuilder();
                        modes.append("enter mode (");
                        modes.append(listEnumNames(Thermostat.SetpointModeEnum.values()));
                        modes.append("): ");
                        logger(modes.toString());
                        final String modeInput = scanner.nextLine();
                        
                        logger("enter amount (from -12.7 to 12.7 degrees): ");
                        final String amountInput = scanner.nextLine();
                        for (Thermostat.SetpointModeEnum setpoint : Thermostat.SetpointModeEnum.values()) {
                            if (setpoint.name().equalsIgnoreCase(modeInput)) {
                                Double amountNumber;
                                try {
                                    amountNumber = Double.parseDouble(amountInput);
                                } catch (NumberFormatException e) {
                                    logger("invalid number");
                                    continue;
                                }
                               setpointCommand = device.adjustOccupiedSetpoint(setpoint, amountNumber);
                            }
                        }
                        if (setpointCommand == null || setpointCommand.isCompletedExceptionally()) {
                            logger("command did not succeed");
                            continue;
                        } else {
                            setpointCommand.whenComplete((result, error) -> {
                                logger("Command: " + input + " finished");
                                if (error != null) {
                                    logger("error: " + error.getMessage());
                                } else {
                                    logger("result: " + result);
                                }
                            });
                        }
    
                    } else if ("changeFanMode".equalsIgnoreCase(input)) {
                        final StringBuilder modes = new StringBuilder();
                        modes.append("enter mode (");
                        modes.append(listEnumNames(FanControl.FanModeEnum.values()));
                        modes.append("): ");
                        logger(modes.toString());
                        final String modeInput = scanner.nextLine();
                        for (FanControl.FanModeEnum fanMode : FanControl.FanModeEnum.values()) {
                            if (fanMode.name().equalsIgnoreCase(modeInput.toLowerCase())) {
                                command = device.changeFanMode(fanMode);
                                break;
                            }
                        }
    
                    } else if ("readRunningMode".equalsIgnoreCase(input)) {
                        propertyFuture = device.readRunningMode();
    
                    } else if ("readFanMode".equalsIgnoreCase(input)) {
                        propertyFuture = device.readFanMode();
                    
                    } else if ("readOccupiedCoolingSetpoint".equalsIgnoreCase(input)) {
                        propertyFuture = device.readOccupiedCoolingSetpoint();
    
                    } else if ("readOccupiedHeatingSetpoint".equalsIgnoreCase(input)) {
                        propertyFuture = device.readOccupiedHeatingSetpoint();
    
                    } else if ("setOccupiedCoolingSetpoint".equalsIgnoreCase(input)) {
                        logger("enter temperature (i.e. 24.5): ");
                        final String amountInput = scanner.nextLine();
                        Double amountNumber;
                        try {
                            amountNumber = Double.parseDouble(amountInput);
                        } catch (NumberFormatException e) {
                            logger("invalid number");
                            continue;
                        }
                        command = device.setOccupiedCoolingSetpoint(amountNumber);
    
                    } else if ("setOccupiedHeatingSetpoint".equalsIgnoreCase(input)) {
                        logger("enter temperature (i.e. 22.0): ");
                        final String amountInput = scanner.nextLine();
                        Double amountNumber;
                        try {
                            amountNumber = Double.parseDouble(amountInput);
                        } catch (NumberFormatException e) {
                            logger("invalid number");
                            continue;
                        }
                        command = device.setOccupiedHeatingSetpoint(amountNumber);
                    }
                    
                    // handle the command results
                    if (command == null && propertyFuture == null) {
                        logger("command is null");
                        continue;
                    }
                    if (command != null) {
                        Optional<Property> value = command.getCachedProperty();
                        if (value.isPresent()) {
                            logger("command cache result: " + value.get().getValue());
                        }
                        command.getFuture().whenComplete((result, error) -> {
                            logger("Command: " + input + " finished");
                            if (error != null) {
                                error.printStackTrace();
                            } else {
                                logger("result: " + result.getValue());
                            }
                        });
                    } else {
                        propertyFuture.whenComplete((result, error) -> {
                            logger("Command: " + input + " finished");
                            if (error != null) {
                                error.printStackTrace();
                            } else {
                                logger("result: " + result.getValue());
                            }
                        });
                    }
    
                } else if ("readAttribute".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    if ( !(d instanceof ZigbeeDevice) ) { logger("not a zigbee device"); continue; }
                    
                    logger("Enter cluster id as hex string (i.e: 0x0000): ");
                    final String clusterIdStr = scanner.nextLine();
                    final int clusterId = JSONUtils.parseUInt16(clusterIdStr).getValue();
    
                    logger("Enter attribute id as hex string (i.e. 0x0001): ");
                    final String attrIdStr = scanner.nextLine();
                    final int attributeId = JSONUtils.parseUInt16(attrIdStr).getValue();
    
                    ZigbeeDevice z = (ZigbeeDevice)d;
                    try { 
                        Property p = z.readAttribute(clusterId, attributeId).get(5, TimeUnit.SECONDS);
                        logger(p.toString());
                    } catch (Exception e) {
                        logger("read attr error: " + e.getMessage());
                    }
    
                } else if ("writeAttribute".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    if ( !(d instanceof ZigbeeDevice) ) { logger("not a zigbee device"); continue; }
                    
                    logger("Enter cluster id as hex string (i.e: 0x0000): ");
                    final String clusterIdStr = scanner.nextLine();
                    final int clusterId = JSONUtils.parseUInt16(clusterIdStr).getValue();
    
                    logger("Enter attribute id as hex string (i.e. 0x0001): ");
                    final String attrIdStr = scanner.nextLine();
                    final int attributeId = JSONUtils.parseUInt16(attrIdStr).getValue();
                    
                    logger("Enter data type as hex string (i.e. bool:0x10, uint8:0x21, uint16:0x22, int8:0x28, int16:0x29): ");
                    final String dataTypeStr = scanner.nextLine();
                    final short dataType = JSONUtils.parseUInt8(dataTypeStr).getValue();
                    
                    logger("Enter value as byte array (i.e. entered as LSB, so 2700 would be 8c 0a ): ");
                    final String valueStr = scanner.nextLine();
                    final byte[] value = SerialUtil.hexStringToLSBByteArray(valueStr.replaceAll(" ",""));
    
                    try {
                        ZigbeeDevice z = (ZigbeeDevice)d;
                        Property p = z.writeAttribute(clusterId, attributeId, dataType, value).get(5, TimeUnit.SECONDS);
                        logger(p.toString());
                    } catch (Exception e) {
                        logger("write attr error: " + e.getMessage());
                    }
    
                } else if ("sendZclCommand".equalsIgnoreCase(input)) {
                    logger("Enter a device id: ");
                    final String idStr = scanner.nextLine();
                    final Device d = gw.getDevice(idStr);
                    if (d == null) { logger("Device not found"); continue; }
                    if ( !(d instanceof ZigbeeDevice) ) { logger("not a zigbee device"); continue; }
                    
                    logger("Enter command id as hex string (i.e. 0x01): ");
                    final String cmdIdStr = scanner.nextLine();
                    final short commandId = (short) JSONUtils.parseUInt8(cmdIdStr).getValue();
                    
                    logger("Enter cluster id as hex string (i.e: 0x0000): ");
                    final String clusterIdStr = scanner.nextLine();
                    final int clusterId = JSONUtils.parseUInt16(clusterIdStr).getValue();
    
                    logger("Enter the payload as byte array, or enter for none (i.e. each field in the payload is entered as LSB, so 2700 would be 8c 0a ): ");
                    final String payloadStr = scanner.nextLine();
                    final byte[] payload = SerialUtil.hexStringToLSBByteArray(payloadStr.replaceAll(" ",""));
                    
                    logger("Is it cluster specific? (true or false): ");
                    final String clusterSpecificStr = scanner.nextLine();
                    final boolean clusterSpecific = (clusterSpecificStr.equalsIgnoreCase("true"));
                    
                    logger("Does command require a response? (true or false): ");
                    final String requiresResponseStr = scanner.nextLine();
                    final boolean requiresResponse = (requiresResponseStr.equalsIgnoreCase("true"));
    
                    try {
                        ZigbeeDevice z = (ZigbeeDevice)d;
                        Property p = z.sendZclCommand(commandId, payload, clusterId, 0, true, clusterSpecific, requiresResponse).get(5, TimeUnit.SECONDS);
                        logger(p.toString());
                    } catch (Exception e) {
                        logger("send zcl command error: " + e.getMessage());
                    }
    
                } else {
                    logger("Unknown command; try h for a list of commands");
                }
            } catch (Exception ex) {
                ex.printStackTrace();
                logger(ex.toString());
            }
        }
    }

    @Override
    public void onDeviceEvent(DeviceEvent event) {
        final Device d = event.getDevice();
        final DeviceEventStatus eventStatus = event.getStatus(); 
        
        printDeviceEvent(event.getStatus(), d.getID(), d.getDeviceType().getType(), d.getDeviceType().getCategory());
        
        switch (eventStatus) {
        case DEVICE_ADDED:
            d.addPropertyUpdateHandler( (device, property) -> {
                logger("Property update triggered for device: " + device.getID());
                logger(property.toString());
            });
            break;
        default:
            break;
        }
    }
}
