 package gatewayapi.sample.iaszone;

import java.util.Optional;

import com.google.gson.JsonObject;
import com.google.gson.JsonParser;
import com.mmbnetworks.gatewayapi.PropertyCommandData;
import com.mmbnetworks.gatewayapi.entity.Device;
import com.mmbnetworks.gatewayapi.entity.Property;
import com.mmbnetworks.gatewayapi.event.device.DeviceEvent;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventHandler;
import com.mmbnetworks.gatewayapi.event.device.DeviceEventStatus;
import com.mmbnetworks.gatewayapi.exception.InvalidInputException;
import com.mmbnetworks.rotarrandevicemodel.jsonsupport.PropertyNames;
import com.mmbnetworks.rotarrandevicemodel.zigbee.property.IASCIEAddressProperty;
import com.mmbnetworks.rotarrandevicemodel.zigbee.property.ZoneStatusProperty;

public class IasDeviceEventHandler implements DeviceEventHandler {

    private final IasPropertyHandler propertyHandler;
    private final IasProtocolPassthroughHandler protocolHandler;
    
    public IasDeviceEventHandler() {
        this.propertyHandler = new IasPropertyHandler();
        this.protocolHandler = new IasProtocolPassthroughHandler();
    }

    @Override
    public void onDeviceEvent(DeviceEvent event) {
        // This method gets called when a device is added to, or removed from, the network.
        // By default, client callbacks are called from a single thread managed by GAPI.
        // Therefore, avoid writing blocking code in this method, because other client callbacks
        // (such as PropertyUpdateHandlers, or ProtocolHandlers) may be blocked.
        
        final DeviceEventStatus eventStatus = event.getStatus();
        final Device device = event.getDevice();
        final String deviceId = device.getID();
        final String deviceType = device.getDeviceType().getType();
        final String deviceCategory = device.getDeviceType().getCategory();
        
        printDeviceEvent(eventStatus, deviceId, deviceType, deviceCategory);
        
        switch (eventStatus) {
        
        case DEVICE_ADDED:
            // Register a property handler, to be notified when a Property on the Device changes.
            device.addPropertyUpdateHandler(propertyHandler);
            
            // Register a protocol passthru handler, to be notified about low-level
            // protocol-specific messages (i.e. zigbee messages from specific clusters)
            device.addProtocolHandler(protocolHandler);

            // If we find an iasCieAddress property, then write to it with the EUI of the gateway.
            device.getProperty(IASCIEAddressProperty.NAME).thenAccept( property -> {
                printDiscoveredProperty(property);
                handleCieAddressRegistration(device, property);
            });
            
            // If we find a zoneStatus property, then bind to it.
            device.getProperty(ZoneStatusProperty.NAME).thenAccept( property -> {
                printDiscoveredProperty(property);
                handleZoneStatusBinding(device, property);
            });

            break;

        case DEVICE_REMOVED:
            device.removeAllProtocolHandlers();
            device.removeAllPropertyUpdateHandlers();
            break;

        default:
            break;
        }
    }
    
    private static void printDeviceEvent(DeviceEventStatus eventStatus, String deviceId, String deviceType, String deviceCategory) {
        StringBuilder sb = new StringBuilder();
        sb.append("Device Event: ").append(eventStatus.toString()).append(", ")
          .append("Device Id: ").append(deviceId).append(", ")
          .append("Device Type: ").append(deviceType).append(", ")
          .append("Device Category: ").append(deviceCategory);
        System.out.println(sb.toString());
    }

    private static void printDiscoveredProperty(Property p) {
        System.out.println("Discovered: " + p.toString());
    }
    
    private static void handleCieAddressRegistration(final Device device, final Property property) {
        // IAS Zone devices require that the address of the gateway be written to one of its attributes
        // before the device will function, so we do that here
        if (property.getName().equalsIgnoreCase(IASCIEAddressProperty.NAME)) {
            final String gatewayEui = getConnectedGatewayEui(device);
            if ( !gatewayEui.isEmpty() && !property.getValue().equalsIgnoreCase(gatewayEui) ) {
                writeCieAddress(device, gatewayEui);
            }
        }
    }
    
    private static void handleZoneStatusBinding(final Device device, final Property property) {
        if (property.getName().equalsIgnoreCase(ZoneStatusProperty.NAME)) {
            // do a binding on the ZoneStatus property, so our PropertyUpdateHandler will trigger
            // when the ZoneStatus changes.
            try {
                device.bindDevice(device.getConnectedGatewayId().get(), ZoneStatusProperty.NAME)
                      .thenAccept(resultStr -> System.out.println("binding result: " + resultStr));

            } catch (InvalidInputException e) {
                System.out.println("error performing zone status binding");
            }   
        }
    }

    private static String getConnectedGatewayEui(final Device device) {
        // Every Device stores protocol-specific attributes (like a zigbee device's EUI), in a Property object
        // called "ProtocolProperty".
        // This code demonstrates how to grab the ProtocolProperty of the gateway that a device is connected to,
        // and then grab the EUI out of it.
        // It will return the EUI of the connected gateway if it exists, or an empty string if an error occurred.
        String gatewayEui = "";
        try {
            Optional<Property> o = IasZoneApp.gw.getDevice(device.getConnectedGatewayId().get()).getCachedProperty(PropertyNames.PROTOCOL_PROPERTY_NAME);
            if (o.isPresent()) {
                final JsonParser parser = new JsonParser();
                final JsonObject jsonProperty = parser.parse(o.get().getValue()).getAsJsonObject();
                gatewayEui = jsonProperty.get("eui").getAsString();
            }
        } catch (Exception ex) {
            gatewayEui = "";    
        }
        return gatewayEui;
    }

    private static PropertyCommandData writeCieAddress(final Device device, final String eui) {
        // this is a non-blocking call to update the named property on a remote device
        System.out.println("Writing CIE Address: " + eui);
        return device.updateProperty(new Property(IASCIEAddressProperty.NAME, IASCIEAddressProperty.TYPE, eui));
    }

}
